#include "knaps.hpp"
#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include <iostream.h>

#define min(x,y) ((x)<(y)?(x):(y))
#define max(x,y) ((x)>(y)?(x):(y))
#define RANDOM (((double)rand())/((double)RAND_MAX+1))

#define MAX_CHANGE 0.2                          // Maximaler Anteil geaenderter Chromosomen bei PMX
#define MAXCONTAINERS 100                       // Maximale Anzahl (+1) an Containern vor Abbruch



// ---------- Elementfunktionen von CHROMOSOME ---------------------------

// Austausch zweier Gene im Chromosom:
//  Parameter:  iPos1   -> Erstes zu vertauschendes Gen
//              iPos2   -> Zweites zu vertauschendes Gen
//  Rueckgabe:  Referenz auf das geaenderte Chromosom
CHROMOSOME & CHROMOSOME::Exchange(const NAT1 &iPos1, const NAT1 &iPos2)
{
    NAT1 iBackup1=(*this)[iPos1];

    assert(iPos1<=len() && iPos2<=len());
    ovwrt1(iPos1,(*this)[iPos2]);
    ovwrt1(iPos2,iBackup1);

    return *this;
}

// Bestimmt die Position eines Gens im Chromosom:
//  Parameter:  Gen     -> Gesuchtes Gen
//  Rueckgabe:  0 wenn das Gen nicht gefunden wurde, sonst die Position,
//              beginnend mit 1
NAT0 CHROMOSOME::Position(const NAT1 &Gen)
{
    NAT0 i;

    for(i=len().cv_nat0();i>=1 && (*this)[i.cv_nat0()]!=Gen;--i) ;

    return i;
}



// ---------- Elementfunktionen von POPULATION ---------------------------

// Auswahl eines Individuums fuer Kreuzung:
// Es wird aus der alten Population <oldpop> ausgewaehlt.
// Die Wahrscheinlichkeit ausgewaehlt zu werden ist proportional zur Fitness.
//  Rueckgabe:   Nr. des Individuums, beginnend bei 1
NAT1 POPULATION::Select()
{
    REAL dRnd=(REAL)RANDOM,dFit=sumfitness()*dRnd;
    int nr;

    nr=(int)(dRnd*popsize().cv_nat1()).cv_real();
    while(nr<(int)popsize().cv_nat1() && ind()[nr+1].fitness()<dFit)
        nr+=(int)step().cv_nat1();
    nr=min(nr,(int)popsize().cv_nat1()-1);
    while(nr>=0 && ind()[nr+1].fitness()>dFit)
        nr-=(int)step().cv_nat1();
    for(nr=max(0,nr);                                 // ..und in kleinen Schritten
        nr<(int)popsize().cv_nat1() && ind()[nr+1].fitness()<dFit;++nr) ;       // vorwaerts suchen
    
    return nr+1;
}

// Scaling der Fitness-Werte:
// Vorheriger Aufruf dieser Funktion ist Voraussetzung fuer den Aufruf der
// Funktion 'Select' !
// Alle werden modifiziert mit f'=a*(f-b)+b, wobei b der Durchschnitt der
// Fitness-Werte ist und a so gewaehlt wird, dass fuer den Besten Wert gilt:
// f'=2*b. Werte von f' unter 0 werden zu 0.
// Danach werden die Fitness-Werte summiert und die aktuelle SUMME bis zur
// bisherigen Position im Fitness-Wert gespeichert (=> Summierung muss nicht
// mehr jedesmal beim Aufruf der 'Select'-Funktion stattfinden.
//  Rueckgabe:  Referenz auf die geaenderte Population
POPULATION & POPULATION::Scaling()
{
    REAL b=sumfitness()/popsize().cv_nat1(),    // s.o.
        a,
        dFit;

    if(bestf()!=b) a=b/(bestf()-b); else a=1;
    chg_sumfitness(0);
    for(NAT1 i=1;i<=popsize();++i)              // Alle Fitness-Werte reduzieren
    {
        dFit=a*(ind()[i].fitness()-b)+b;        // Neuen Fitness-Wert berechnen
        chg_sumfitness(sumfitness()+                    // Summe bilden
            (dFit.cv_real()>=0?dFit.cv_real():0));
        chg_ind(ind().ovwrt1(i,ind()[i].chg_fitness(sumfitness())));    // und im Individuum speichern
    }

    return *this;
}

// Summe der Fitness-Werte der Population ermitteln:
// Vorheriger Aufruf dieser Funktion ist Voraussetzung fuer ordnungsgemaessen
// Ablauf der Funktion 'Scaling' !
// Ermittelt wird: Summer der Fitness-Werte, Nr und Fitness-Wert des Besten
//                 und schlechtesten Chromosoms
// Diese Werte werden in der Population eingetragen.
//  Rueckgabe:  Referenz auf die geaenderte Population
POPULATION & POPULATION::MakeSumFitness()
{
    REAL dF;

    chg_bestf(ind()[1].fitness());              // Besten und schlechtesten auf
    chg_worstf(ind()[1].fitness());             //  1. Element festlegen
    chg_best(1);
    chg_worst(1);
    chg_sumfitness(0);

    for(NAT1 i=1;i<=popsize();++i)              // Gesamte Population abarbeiten
    {
        dF=ind()[i].fitness();
        chg_sumfitness(dF+sumfitness());        // Fitness zur Summe addieren
        if(dF>bestf())                          // Besser als Bester ?
        {
            chg_best(i);                        // Ja, dann neuer Bester
            chg_bestf(dF);
        }
        else
            if(dF<worstf())                     // Schlechter als Schlechtester
            {
                chg_worst(i);                   // Ja, dann neuer Schlechtester
                chg_worstf(dF);
            }
    }

    return *this;
}


// ---------- Elementfunktionen von GA ---------------------------------

// Initialisieren u. Anfangspopulation erzeugen:
//  Parameter:  c       -> Benutzer Container
//              b       -> Benutzte Bloecke
//              iPS     -> Groesse der Population (muss gerade sein und >0 !)
//              iSL     -> Anzahl der Genen im Chromosom (>1 !)
//              dPC     -> Wahrscheinlichkeit fuer Kreuzung ([0-1[ !)
//              dPM     -> Wahrscheinlichkeit fuer Mutation ([0-1[ !)
//              dMC     -> Maximale Aenderung bei PMX ([0-1[!)
//  Rueckgabe:  Referenz auf geaendertes GA
GA & GA::Init(const CONTAINER &c, const BLOCKS &b, const NAT1 &iPS,
              const NAT1 &iSL, const REAL &dPC, const REAL &dPM,
              const REAL &dMC)
{
    assert(iPS%2==0 && iPS>0 && iSL>1 && dPC>=0 && dPC<1 && dPM>=0 && dPM<1 &&
           dMC>=0 && dMC<1);

    *this=GA(iPS,iSL,                           // Initialiserung leerer GA
             POPULATION(INDIVIDUALS(empty),iPS,iSL,0,0,0,0,0,0),
             POPULATION(INDIVIDUALS(empty),iPS,iSL,0,0,0,0,0,0),
             dPC,dPM,dMC,0,c,b,0);

    for(NAT1 i=1;i<=iPS;++i)                    // Chromosomen + Fitness erzeugen
    {
        CHROMOSOME Chrom=MakeChrom(iSL);
        INDIVIDUALS newInds(oldpop().ind()+INDIVIDUAL(Chrom,GetFitness(Chrom)));

        chg_oldpop(oldpop().chg_ind(newInds));
    }
    chg_oldpop(oldpop().MakeSumFitness());      // Fitness-Werte berechnen

    return *this;
}

// Naechste Generation einer Population erzeugen:
// Population wird skaliert, danach werden aus je zwei alten Individuen zwei
// neue erzeugt und in die neue Population eingefuegt bis diese voll ist.
//  Rueckgabe:  Referenz auf geaenderten GA
GA & GA::NextGen()
{
    INDIVIDUAL oBest=oldpop().ind()[oldpop().best()],
               oInd1,oInd2;
    NAT1 iInd1,iInd2;

    chg_oldpop(oldpop().Scaling().              // Fitness-Werte skalieren und Schrittweite festlegen
                chg_step((int)sqrt((double)popsize().cv_nat1())));
    chg_newpop(POPULATION(INDIVIDUALS(empty),popsize(),stringlen(),0,0,0,0,0,0));

    for(NAT1 i=1;i<=popsize();i+=2)             // Population auffuellen
    {
        for(iInd1=oldpop().Select();            // Zwei Individuen auswaehlen
            (iInd2=oldpop().Select())==iInd1;) ;
        oInd1=Crossover(oldpop().ind()[iInd1],oldpop().ind()[iInd2],  // ..und kreuzen
                        oInd2);
        chg_newpop(newpop().chg_ind(newpop().ind()+oInd1+oInd2));      // Beide zur Population hinzufuegen
    }
    chg_oldpop(newpop());                       // Neue Generation wird zur Alten
    chg_oldpop(oldpop().MakeSumFitness());      // Gesamtfitness berechnen
    if(oldpop().bestf()<oBest.fitness())        // Bester Neuer schlechter als bester Alter
        chg_oldpop(oldpop().chg_ind(oldpop().ind().ovwrt1(1,oBest)));  // => Bester der Alten ueberlebt
    chg_generation(generation()+1);
    chg_oldpop(oldpop().MakeSumFitness());      // Gesamtfitness berechnen

    return *this;
}

// Zwei Eltern-Chromosomen kreuzen und daraus zwei Kind-Chromosomen erstellen:
// Kreuzung findet mit PMX-Operator statt.
//  Parameter:  Parent1 -> Erstes zu kreuzendes Chromosom
//              Parent2 -> Zweites zu kreuzendes Chromosomen
//              Child2  -> Referenz auf zweites Kind
//  Rueckgabe:  Zweites Kind
INDIVIDUAL GA::Crossover(const INDIVIDUAL &Parent1, const INDIVIDUAL &Parent2,
                         INDIVIDUAL &Child2)
{
    INDIVIDUAL Child1=Parent1;                  // Eltern in Kinder kopieren

    Child2=Parent2;
    Child1.chg_chrom(Mutation(Child1.chrom())); //  und mutieren
    Child2.chg_chrom(Mutation(Child2.chrom()));
    if(RANDOM<pcross().cv_real())               // Eltern kreuzen ?
    {                                           // Ja, ab ins Bett...
        NAT0 i,
             iRange,                            // Anzahl der zu kreuzenden Gene
             from,to;                           // Erstes und letztes(+1) Gen

        for(iRange=max((int)(RANDOM*stringlen().cv_nat1()*maxchange().cv_real()),1);
            (from=(int)(RANDOM*stringlen().cv_nat1()))+iRange>stringlen();) ;
        
        for(i=from,to=from+iRange;i<to;++i)     // Chromosomen mischen nach PMX
        {
            NAT0 iGen11=(Child1.chrom()[i+1]).cv_nat1(), // Die Gene an dieser Position
                 iGen21=(Child2.chrom()[i+1]).cv_nat1(),iGen12,iGen22;

            if(iGen11!=iGen21)                  // Gene sind unterschiedlich ?
            {                                   // Ja, Gen des Partners im
                if((iGen12=Child1.chrom().Position(iGen21.cv_nat0()))!=0)           // Chromosom
                    Child1.chg_chrom(Child1.chrom().Exchange(i+1,iGen12.cv_nat0()));// suchen und
                if((iGen22=Child2.chrom().Position(iGen11.cv_nat0()))!=0)           // vertauschen
                    Child2.chg_chrom(Child2.chrom().Exchange(i+1,iGen22.cv_nat0()));
            }
        }

    }
    Child1.chg_fitness(GetFitness(Child1.chrom())); // Fitness-Werte der Kinder ermitteln
    Child2.chg_fitness(GetFitness(Child2.chrom()));

    return Child1;
}

// Fitness-Wert eines Chromosoms berechnen:
// Ausser der Berechnung der Fitness werden die <x>- und <y>-Werte
// der Blocke werden entsprechend der Anordnung der Bloecke modifiziert.
// Fitness wird berechnet als:
// AnzahlDerBloecke*Containerbreite - PositionRechterRandDesLetztenBlocks +
// FreierPlatzImContainer/ContainerFlaeche
//  Parameter:  Chrom   -> Chromosom, dessen Fitness zu ermitteln ist
//  Rueckgabe:  Fitness-Wert des Chromsoms
REAL GA::GetFitness(const CHROMOSOME &Chrom)
{
    CONFILL Filled(0,con().cheight());          // Leerer Container
    NAT1 i;
    NAT0 j,
         iMaxW=0,                               // Hoechste Fuellung im aktuellen Container
         iLeft=con().cwidth()*con().cheight(),  // Freie Felder im Container
         x=0,y=0,                               // Koordinaten fuer naechsten Block
         WorstFill=con().cwidth()*stringlen();  // Schlechteste Fuellung: Nur ein Block pro Container
    REAL fit;

    chg_containers(1);                          // Mit einem Container anfangen
    for(i=1;i<=stringlen();++i)                 // Alle Bloecke bearbeiten
    {
        BLOCK b=bl()[Chrom[i]];
        int found=false;

        if(y+b.bheight()>con().cheight()) y=0;  // Block zu hoch, dann von oben
        
        x=Filled[y+1];                          // "Links" im Container anfangen
        while(!found && x+b.bwidth()<=con().cwidth())   // Position ohne Ueberschneidung
        {                                       //  mit anderen Bloecken suchen
            found=true;
            for(j=y;found && j<y+b.bheight();++j)
                found=(Filled[j+1]<=x);
            if(!found) ++x;
        }
        if(x+b.bwidth()>con().cwidth())         // Container uebergelaufen ?
        {
            chg_containers(containers()+1);     // Ja, dann Neuen beginnen
            assert(containers()<MAXCONTAINERS); // Dann liegt wohl ein Fehler vor ???
            iMaxW=0;x=0;y=0;
            Filled=CONFILL(0,con().cheight());  // Container wieder voll
            iLeft=con().cwidth()*con().cheight();
        }

        b.chg_x(x+(containers()-1)*con().cwidth()); // Position des Blockes speichern
        b.chg_y(y);
        for(j=y;j<y+b.bheight();++j)
            Filled.ovwrt1(j+1,x+b.bwidth());
        y+=b.bheight();                         // Naechsten Block darunter Positionieren
        iMaxW=max(x+b.bwidth(),iMaxW);          // Belegte Breite messen
        iLeft-=b.bwidth()*b.bheight();          // Block von freier Flaeche abziehen
        chg_bl(bl().ovwrt1(Chrom[i],b));
    }

    fit=/*(double)iLeft.cv_nat0()/(con().cheight()*con().cwidth()).cv_nat0()+*/
        (WorstFill-iMaxW-(containers()-1)*con().cwidth()).cv_nat0();

    if(fit>0) return fit; else return 0;        // Nur Positive Fitness zurueckgeben 
}

// Eventuelle Mutation eines Chromosoms:
// Alle Gene eines Chromosoms werden abhaengig vom Zufall (Wahrscheinlichkeit
// <oChrom>) mutiert.
//  Parameter:  Zu mutierendes Chromosom
//  Rueckgabe:  Eventuell mutiertes Chromosom
CHROMOSOME GA::Mutation(const CHROMOSOME &oChrom)
{
    CHROMOSOME chrom=oChrom;

    for(NAT1 i=1;i<=chrom.len();++i)           // Alle Gene pruefen
        if(RANDOM<pmutation().cv_real())        // Mutation dieses Gens ?
            chrom.Exchange(i,(int)(RANDOM*chrom.len().cv_nat0())+1);

    return chrom;
}

// Ein Chromosom erzeugen:
// Es wird ein String erzeugt, der die Zahlen von 0-<iStringLen> in
// zufaelliger Reihenfolge enthaelt.
//  Parameter:  Laenge des zu erzeugenden Chromosomen-Strings
//  Rueckgabe:  Erzeugtes Chromosom
CHROMOSOME GA::MakeChrom(const NAT1 &iStringLen)
{
    CHROMOSOME oChrom(empty),oStd(empty);
    NAT1 iRestLen=iStringLen,
         i;

    for(i=1;i<=iStringLen;++i)                  // Standardstring initialisieren
        oStd+=i;

    for(i=1;i<=iStringLen;++i,--iRestLen)       //  und zufaellig nach <oChrom> uebertragen
    {
        NAT1 j=(int)(RANDOM*iRestLen.cv_nat1())+1;

        oChrom+=oStd[j];                        // Kopie eines Gens
        oStd.ovwrt1(j,oStd[iRestLen]);          // Und Gen aus <Std> "entfernen"
    }

    return oChrom;
}

